ENABLE_LOGGING = false

local rusty_locale = require '__rusty-locale__.locale'
local rusty_icons = require '__rusty-locale__.icons'
local rusty_recipes = require '__rusty-locale__.recipes'

-- Add categories here with max number of inputs and outputs
MIRROR_CATEGORIES = {
  ["oil-processing"] = {
    INPUTS = 2,
    OUTPUTS = 3
  },
  ["chemistry"] = {
    INPUTS = 2,
    OUTPUTS = 2
  }
}

-- Make sure each category above also has H_SHIFT set
-- This isn't automated because display order depends on the order in which processing
-- building fluixboxes are defined - note that refinery is right-to-left and chem plant is
-- left-to-right
MIRROR_ICON = {
  SCALE = 0.25,
  H_SHIFT = {
    ["oil-processing"] = {
      ["input"] = {[1] = 8, [2] = -8},
      ["output"] = {[1] = 12, [2] = 0, [3] = -12}
    },
    ["chemistry"] = {
      ["input"] = {[1] = -10, [2] = 10},
      ["output"] = {[1] = -8, [2] = 8}
    }
  },
  V_SHIFT = {
    ["input"] = -10,
    ["output"] = 12
  }
}

-- Name constants
MIRROR_NAME = {
  GROUP = "mmoto",
  PREFIX = "mirrored",
  SUFFIX = {
    RECIPE = "-mirrored",
    INPUT = "-in",
    OUTPUT = "-out",
    BOTH = "-both"
  }
}

-- This is automatically populated, leave empty
MIRROR_RECIPES = { }

-- Returns true if this recipe has more than one fluid ingredient, or if it has at least
-- one fluid ingredient with an indexed position
function check_inputs(recipe)
  local fluid_count = 0
  if recipe.ingredients then
    for i,ingredient in pairs(recipe.ingredients) do
      if ingredient.type == "fluid" then
        fluid_count = fluid_count + 1
        if ingredient.fluidbox_index then
          fluid_count = fluid_count + 1
        end
      end
    end
  end
  return fluid_count > 1
end

-- Returns true if this recipe has more than one fluid result, or if it has at least one
-- fluid result with an indexed position
function check_outputs(recipe)
  local fluid_count = 0
  if recipe.results then
    for i,result in pairs(recipe.results) do
      if result.type == "fluid" then
        fluid_count = fluid_count + 1
        if result.fluidbox_index then
          fluid_count = fluid_count + 1
        end
      end
    end
  end
  return fluid_count > 1
end

-- Finds all technologies that unlock this recipe and returns a table of their names
function find_technologies(recipe)
  techs = {}
  for i,tech in pairs(data.raw.technology) do
    if tech.effects then
      for j,effect in pairs (tech.effects) do
        if effect.type == "unlock-recipe" and effect.recipe == recipe.name then
          if ENABLE_LOGGING then
            log("Found technology " .. tech.name .. " to unlock recipe " .. recipe.name)
          end
          table.insert(techs, tech.name)
        end
      end
    end
  end
  return techs
end

-- Do productivity modules support this recipe?
function productivity_support(recipe)
  for i,mod in pairs(data.raw.module) do
    if mod.effect.productivity and mod.limitation then
      for j,limit in pairs(mod.limitation) do
        if limit == recipe.name then
          if ENABLE_LOGGING then
            log("Recipe " .. recipe.name .. " compatible with productivity")
          end
          return true
        end
      end
    end
  end
  return false
end

-- Assigns consecutive indices to each fluid ingredient
-- Note: this functions assumes either all fluids have an index already or none do
function normalise_ingredients(recipe)
  for i,ingredient in pairs (recipe.ingredients) do
    if ingredient.fluidbox_index then
      if ENABLE_LOGGING then
        log("Recipe " .. recipe.name .. " ingredients already indexed")
      end
      return
    end
  end
  local index = 1
  for i,ingredient in pairs (recipe.ingredients) do
    if ingredient.type == "fluid" then
      ingredient.fluidbox_index = index
      if ENABLE_LOGGING then
        log("Recipe " .. recipe.name .. " ingredient " .. ingredient.name .. " index " .. index)
      end
      index = index + 1
    end
  end
end

-- Assigns consecutive indices to each fluid result
-- Note: this functions assumes either all fluids have an index already or none do
function normalise_results(recipe)
  for i,result in pairs (recipe.results) do
    if result.fluidbox_index then
      if ENABLE_LOGGING then
        log("Recipe " .. recipe.name .. " results already indexed")
      end
      return false
    end
  end
  local index = 1
  for i,result in pairs (recipe.results) do
    if result.type == "fluid" then
      result.fluidbox_index = index
      if ENABLE_LOGGING then
        log("Recipe " .. recipe.name .. " result " .. result.name .. " index " .. index)
      end
      index = index + 1
    end
  end
end

-- Creates a localised name for this recipe based on the constants defined above
function get_localised_name(recipe, mirror_inputs, mirror_outputs)
  local locale_string = MIRROR_NAME.GROUP .. "." .. MIRROR_NAME.PREFIX
  if mirror_inputs and mirror_outputs then
    locale_string = locale_string .. MIRROR_NAME.SUFFIX.BOTH
  elseif mirror_inputs then
    locale_string = locale_string .. MIRROR_NAME.SUFFIX.INPUT
  elseif mirror_outputs then
    locale_string = locale_string .. MIRROR_NAME.SUFFIX.OUTPUT
  end
  return {locale_string, rusty_locale.of_recipe(recipe).name}
end

-- Gets order of this recipe (or of its product)
function get_recipe_order(recipe)
  if recipe.order then
    return recipe.order
  else
    local product = rusty_recipes.get_main_product(recipe)
    return data.raw[product.type][product.name].order
  end
end

-- Adds a small fluid icon to a recipe's icon with given position
function add_fluid_icon(fluid_name, category, in_out, index, recipe)
  local icons = rusty_icons.of(fluid_name, "fluid")
  for i,icon in pairs(icons) do
    icon.scale = MIRROR_ICON.SCALE
    icon.shift = {
      MIRROR_ICON.H_SHIFT[category][in_out][index],
      MIRROR_ICON.V_SHIFT[in_out]
    }
    table.insert(recipe.icons, icon)
  end
end

-- Create the icon for this recipe by adding an icon for each fluid ingredient/result with
-- an index, i.e. all fluid ingredients/results that can be mirrored
function generate_icon(recipe)
  for i,ingredient in pairs(recipe.ingredients) do
    if ingredient.type == "fluid" and ingredient.fluidbox_index then
      add_fluid_icon(ingredient.name, recipe.category, "input", ingredient.fluidbox_index, recipe)
    end
  end
  for i,result in pairs(recipe.results) do
    if result.type == "fluid" and result.fluidbox_index then
      add_fluid_icon(result.name, recipe.category, "output", result.fluidbox_index, recipe)
    end
  end
end

-- Create a mirrored version of a recipe and add unlocks to appropriate technologies
function create_mirror(recipe, mirror_inputs, mirror_outputs)
  local new_recipe = table.deepcopy(recipe)
  new_recipe.localised_name = get_localised_name(recipe, mirror_inputs, mirror_outputs)
  new_recipe.name = new_recipe.name .. MIRROR_NAME.SUFFIX.RECIPE
  new_recipe.order = get_recipe_order(recipe) .. MIRROR_NAME.SUFFIX.RECIPE
  if mirror_inputs then
    for i,ingredient in pairs(new_recipe.ingredients) do
      if ingredient.type == "fluid" and ingredient.fluidbox_index then
        ingredient.fluidbox_index = MIRROR_CATEGORIES[new_recipe.category].INPUTS - ingredient.fluidbox_index + 1
      end
    end
  end
  if mirror_outputs then
    for i,result in pairs(new_recipe.results) do
      if result.type == "fluid" and result.fluidbox_index then
        result.fluidbox_index = MIRROR_CATEGORIES[new_recipe.category].OUTPUTS - result.fluidbox_index + 1
      end
    end
  end
  if mirror_inputs and mirror_outputs then
    new_recipe.name = new_recipe.name .. MIRROR_NAME.SUFFIX.BOTH
    new_recipe.order = new_recipe.order .. MIRROR_NAME.SUFFIX.BOTH
  elseif mirror_inputs then
    new_recipe.name = new_recipe.name .. MIRROR_NAME.SUFFIX.INPUT
    new_recipe.order = new_recipe.order .. MIRROR_NAME.SUFFIX.INPUT
  elseif mirror_outputs then
    new_recipe.name = new_recipe.name .. MIRROR_NAME.SUFFIX.OUTPUT
    new_recipe.order = new_recipe.order .. MIRROR_NAME.SUFFIX.OUTPUT
  end
  generate_icon(new_recipe)
  data:extend({new_recipe})
  if ENABLE_LOGGING then
    log("Added recipe " .. new_recipe.name)
  end
  -- unlock techs
  for i,tech in pairs(MIRROR_RECIPES[recipe.category][recipe.name].techs) do
    table.insert(
    data.raw.technology[tech].effects,
    {type = "unlock-recipe", recipe = new_recipe.name})
    if ENABLE_LOGGING then
      log("Added " .. new_recipe.name .. " unlock to technology " .. tech)
    end
  end
  -- apply productivity effect
  if MIRROR_RECIPES[recipe.category][recipe.name].productivity then
    for i,mod in pairs(data.raw.module) do
      if mod.effect.productivity and mod.limitation then
        if ENABLE_LOGGING then
          log("Added " .. new_recipe.name .. " productivity effect to " .. mod.name)
        end
        table.insert(mod.limitation, new_recipe.name)
      end
    end
  end
end

-- Look for any recipe that matches the categories defined at the top of this file and
-- check if it has fluids that can be mirrored, as well as any techs that unlock it
for i,recipe in pairs(data.raw.recipe) do
  if MIRROR_CATEGORIES[recipe.category] then
    MIRROR_RECIPES[recipe.category] = MIRROR_RECIPES[recipe.category] or { }
    local mirror_inputs = check_inputs(recipe)
    local mirror_outputs = check_outputs(recipe)
    if mirror_inputs or mirror_outputs then
      if ENABLE_LOGGING then
        log("Found recipe " .. recipe.name .. " from category " .. recipe.category)
      end
      MIRROR_RECIPES[recipe.category][recipe.name] = {
        inputs = mirror_inputs,
        outputs = mirror_outputs,
        techs = find_technologies(recipe),
        productivity = productivity_support(recipe)
      }
    end
  end
end

-- Build mirror recipes for each recipe we found above
for i,category in pairs(MIRROR_RECIPES) do
  for name,mirror in pairs(category) do
    local recipe = data.raw.recipe[name]
    recipe.icons = rusty_icons.of_recipe(recipe)
    recipe.icon = nil
    if mirror.inputs then
      normalise_ingredients(recipe)
    end
    if mirror.outputs then
      normalise_results(recipe)
    end
    if mirror.inputs then
      create_mirror(recipe, true, false)
    end
    if mirror.outputs then
      create_mirror(recipe, false, true)
    end
    if mirror.inputs and mirror.outputs then
      create_mirror(recipe, true, true)
    end
    generate_icon(recipe)
  end
end